@@ -79,7 +79,6 @@ static char get_signature_char(MVMint16 type_id) {
 /* Sets up a callback, caching the information to avoid duplicate work. */
 static char callback_handler(DCCallback *cb, DCArgs *args, DCValue *result, MVMNativeCallback *data);
 static void * unmarshal_callback(MVMThreadContext *tc, MVMObject *callback, MVMObject *sig_info) {
-    MVMNativeCallbackCacheHead *callback_data_head = NULL;
     MVMNativeCallback **callback_data_handle;
     MVMString          *cuid;
 
@@ -89,13 +88,20 @@ static void * unmarshal_callback(MVMThreadContext *tc, MVMObject *callback, MVMO
     /* Try to locate existing cached callback info. */
     callback = MVM_frame_find_invokee(tc, callback, NULL);
     cuid     = ((MVMCode *)callback)->body.sf->body.cuuid;
-    MVM_HASH_GET(tc, tc->native_callback_cache, cuid, callback_data_head);
 
-    if (!callback_data_head) {
-        callback_data_head = MVM_malloc(sizeof(MVMNativeCallbackCacheHead));
+    if (!tc->native_callback_cache) {
+        tc->native_callback_cache = MVM_fixed_size_alloc(tc, tc->instance->fsa,
+                                                         sizeof (MVMStrHashTable));
+        MVM_str_hash_build(tc, tc->native_callback_cache, sizeof(MVMNativeCallbackCacheHead), 0);
+    }
+
+    MVMNativeCallbackCacheHead *callback_data_head
+        = MVM_str_hash_lvalue_fetch(tc, tc->native_callback_cache, cuid);
+
+    if (!callback_data_head->hash_handle.key) {
+        /* MVM_str_hash_lvalue_fetch created a new entry. Fill it in: */
+        callback_data_head->hash_handle.key = cuid;
         callback_data_head->head = NULL;
-
-        MVM_HASH_BIND(tc, tc->native_callback_cache, cuid, callback_data_head);
     }
 
     callback_data_handle = &(callback_data_head->head);
@@ -317,6 +323,7 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
                 args[i - 1].i64 = dcbArgULongLong(cb_args);
                 break;
             default:
+                MVM_free(args);
                 MVM_telemetry_interval_stop(tc, interval_id, "nativecall callback handler failed");
                 MVM_exception_throw_adhoc(tc,
                     "Internal error: unhandled dyncall callback argument type");
@@ -402,6 +409,7 @@ static char callback_handler(DCCallback *cb, DCArgs *cb_args, DCValue *cb_result
             cb_result->l = MVM_nativecall_unmarshal_ulonglong(tc, res.o);
             break;
         default:
+            MVM_free(args);
             MVM_telemetry_interval_stop(tc, interval_id, "nativecall callback handler failed");
             MVM_exception_throw_adhoc(tc,
                 "Internal error: unhandled dyncall callback return type");
@@ -590,7 +598,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
         }
     }
 
-    MVMROOT2(tc, args, res_type, {
+    MVMROOT3(tc, args, res_type, result, {
         MVM_gc_mark_thread_blocked(tc);
         if (result) {
             /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
@@ -725,6 +733,7 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                     break;
                 }
                 default:
+                    MVM_gc_mark_thread_unblocked(tc);
                     MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke failed");
                     MVM_exception_throw_adhoc(tc, "Internal error: unhandled dyncall return type");
             }
