@@ -105,6 +105,7 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     MVMSpeshBB *bb = ig->entry;
     MVMint32 same_hll = target_sf->body.cu->body.hll_config ==
             inliner->sf->body.cu->body.hll_config;
+    MVMuint32 seen_possible_deopt = 0;
     if (no_inline_info)
         *no_inline_info = NULL;
     while (bb) {
@@ -161,7 +162,9 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 }
             }
 
-            /* Check we don't have too many args for inlining to work out. */
+            /* Check we don't have too many args for inlining to work out,
+             * and that we're not reading args after a possible deopt
+             * instruction, because we can't uninline in such a case. */
             else if (opcode == MVM_OP_sp_getarg_o ||
                     opcode == MVM_OP_sp_getarg_i ||
                     opcode == MVM_OP_sp_getarg_n ||
@@ -170,6 +173,16 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     *no_inline_reason = "too many arguments to inline";
                     return 0;
                 }
+                if (seen_possible_deopt) {
+                    *no_inline_reason = "a deopt may happen before arguments are processed";
+                    return 0;
+                }
+            }
+            else if (opcode == MVM_OP_sp_paramnamesused) {
+                if (seen_possible_deopt) {
+                    *no_inline_reason = "a deopt may happen before arguments are processed";
+                    return 0;
+                }
             }
 
             /* Ext-ops need special care in inter-comp-unit inlines. */
@@ -180,6 +193,10 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     demand_extop(tc, target_cu, source_cu, ins->info);
             }
 
+            /* Record if this instruction would cause us to deopt. */
+            if (ins->info->may_cause_deopt)
+                seen_possible_deopt = 1;
+
             ins = ins->next;
         }
         bb = bb->linear_next;
@@ -303,8 +320,10 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadConte
      * the args specialization). */
     ig = MVM_spesh_graph_create(tc, target_sf, 0, 1);
     MVM_spesh_args_from_callinfo(tc, ig, call_info, type_tuple);
-    MVM_spesh_facts_discover(tc, ig, NULL, 0);
-    MVM_spesh_optimize(tc, ig, NULL);
+    MVMROOT(tc, target_sf, {
+        MVM_spesh_facts_discover(tc, ig, NULL, 0);
+        MVM_spesh_optimize(tc, ig, NULL);
+    });
 
     /* See if it's inlineable; clean up if not. */
     if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason, no_inline_info)) {
@@ -630,6 +649,14 @@ static MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     ins->operands[0].reg.orig += inliner->num_locals;
                 }
             }
+            else if (opcode == MVM_OP_takenextdispatcher) {
+                /* If the inlining code doesn't set a next dispatcher, takenextdispatcher
+                 * in the inlinee can be assumed to always return null. */
+                if (!inliner->sets_nextdispatcher && !inlinee->sets_nextdispatcher) {
+                    ins->info = MVM_op_get_op(MVM_OP_null);
+                    ins->operands[0].reg.orig += inliner->num_locals;
+                }
+            }
             else {
                 if (!same_comp_unit) {
                     if (ins->info->opcode == MVM_OP_const_s) {
@@ -997,6 +1024,7 @@ static void return_to_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *re
                    MVMSpeshIns *return_ins, MVMSpeshOperand target,
                    MVMuint16 box_type_op, MVMuint16 box_op) {
     MVMSpeshOperand type_temp     = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_obj);
+    MVMSpeshFacts *target_facts;
 
     /* Create and insert boxing instruction after current return instruction. */
     MVMSpeshOperand ver_target = MVM_spesh_manipulate_new_version(tc, g, target.reg.orig);
@@ -1008,7 +1036,12 @@ static void return_to_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *re
     box_operands[1]               = return_ins->operands[0];
     box_operands[2]               = type_temp;
     MVM_spesh_manipulate_insert_ins(tc, return_bb, return_ins, box_ins);
-    MVM_spesh_get_facts(tc, g, ver_target)->writer = box_ins;
+    target_facts = MVM_spesh_get_facts(tc, g, ver_target);
+    target_facts->writer = box_ins;
+    target_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE | MVM_SPESH_FACT_CONCRETE | MVM_SPESH_FACT_KNOWN_BOX_SRC;
+    target_facts->type = box_op == MVM_OP_box_i ? g->sf->body.cu->body.hll_config->int_box_type :
+                         box_op == MVM_OP_box_n ? g->sf->body.cu->body.hll_config->num_box_type :
+                                                  g->sf->body.cu->body.hll_config->str_box_type;
     MVM_spesh_usages_add_by_reg(tc, g, box_operands[1], box_ins);
     MVM_spesh_usages_add_by_reg(tc, g, box_operands[2], box_ins);
 
